"""
_constants.py
05. February 2023

really don't know why I put this file in types, but here are a few constants

Author:
Nilusink
"""
# directional
LEFT: str = "left"
RIGHT: str = "right"
TOP: str = "top"
BOTTOM: str = "bottom"

# sticky (no I didn't put in all combinations because you
# can just make them yourself, the order doesn't matter)
N: str = "n"
S: str = "s"
E: str = "e"
W: str = "e"
NS: str = N + S
SW: str = S + W
NSEW: str = N + S + E + W
